from transliterate.base import TranslitLanguagePack, registry

class TgRuLanguagePack(TranslitLanguagePack):
    language_code = "tgru"
    language_name = "tgru"

    character_ranges = ((0x0400, 0x04FF), (0x0500, 0x052F))

    mapping = (
        u"abvgdezijklmnoprstufhcC'y'ABVGDEZIJKLMNOPRSTUFH'Y'",
        u"абвгдезийклмнопрстуфхцЦъыьАБВГДЕЗИЙКЛМНОПРСТУФХЪЫЬ",
    )

    #reversed_specific_mapping = (
    #    u"ъьЪЬ",
    #    u"''''"
    #)

    pre_processor_mapping = {
        u"zh": u"ж",
        "yo": 'ё',
        u"ch": u"ч",
        u"sh": u"ш",
        u"sch": u"щ",
        u"yu": u"ю",
        u"ya": u"я",
        "Yo": 'Ё',
        u"Zh": u"Ж",
        u"Ts": u"Ц",
        u"Ch": u"Ч",
        u"Sh": u"Ш",
        u"Sch": u"Щ",
        u"Yu": u"Ю",
        u"Ja": u"Я",
        u"EH": u"Э",
        u"eh": u"э"
    }


registry.register(TgRuLanguagePack)




#!/usr/bin/env python

"""
    script to check via docker container 
    https://github.com/arnidan/nsfw-api

    manual curl test: curl -X POST http://localhost:3000/classify -F image=@/tmp/eropicture.jpg

"""


import sys
import json
import os
import requests
from nudecrawler.exceptions import *
from nudecrawler.localimage import basic_check
from urllib.parse import urljoin

#start_detector = "sudo docker run --rm --name nsfw-api -d -p 127.0.0.1:5000:5000/tcp --env PORT=5000 eugencepoi/nsfw_api:latest"
start_detector = "sudo docker run --rm --name nsfw-api -d -p 3000:3000 ghcr.io/arnidan/nsfw-api:latest"

timeout = 10

def detect_image(path, address, thresholds, verbose=False):

    req_url = urljoin(address,'/classify')

    try:
        files = {
            'image': open(path, 'rb')
        }
        r = requests.post(req_url, files=files)
        if r.status_code == 500:
            if verbose:                
                print(r.text)
            return  0            

        if r.status_code != 200:
            print(os.getenv('NUDECRAWLER_PAGE_URL'))
            print(os.getenv('NUDECRAWLER_IMAGE_URL'))
            print(r.text)
            print(r.json())

    except requests.Timeout as e:
        # timeout: not interesting image
        print("TIMEOUT")
        sys.exit(0)
    except requests.RequestException as e:
        print(e)
        print("maybe detector not running?")
        print(start_detector)
        print("or add -a to skip filtering")
        sys.exit(100)

    results = r.json()
    rround = {key : round(results[key], 2) for key in results}


    for cls in thresholds:
        if results[cls] > thresholds[cls]:
            if verbose:
                print(f"Nude {path} {cls}: {results[cls]:.2f} > {thresholds[cls]}")
            return 1
    
    if verbose:
        print(f"Safe {path}: {rround}")
    return 0



    # return int(r.json()['neutral'] < threshold)

def main():
    image_path = sys.argv[1]

    thresholds=dict()

    detector_address = os.getenv('DETECTOR_ADDRESS', 'http://localhost:3000/')
    detector_threshold_str = os.getenv('DETECTOR_THRESHOLD', '0.5')
    verbose = bool(os.getenv('DETECTOR_VERBOSE', ""))

    for cls in ['hentai', 'porn', 'sexy']:
        thresholds[cls] = float(os.getenv(f'DETECTOR_THRESHOLD_{cls.upper()}', detector_threshold_str))

    c = detect_image(image_path, address=detector_address, thresholds=thresholds, verbose=verbose)
    sys.exit(c)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
    script to check via docker container 
    https://github.com/open-dating/adult-image-detector
"""

import sys
import os
import requests
from nudecrawler.exceptions import *
from nudecrawler.localimage import basic_check
from urllib.parse import urljoin

start_detector = "sudo docker run --rm --name aid -d -p 9191:9191 opendating/adult-image-detector"

def detect_nudity(path, address, threshold):
    endpoint = urljoin(address, '/api/v1/detect')

    files = {'image': open(path,'rb')}
    try:
        r = requests.post(endpoint, files=files)
        
    except requests.RequestException as e:
        print(e)
        print("maybe detector not running?")
        print(start_detector)
        print("or add -a to skip filtering")
        sys.exit(100)

    # return r.json()['an_algorithm_for_nudity_detection']
    return int(r.json()['open_nsfw_score'] > threshold)

def main():
    image_path = sys.argv[1]

    detector_address = os.getenv('DETECTOR_ADDRESS', 'http://localhost:9191/')
    detector_threshold = float(os.getenv('DETECTOR_THRESHOLD', '0.5'))
    min_w = int(os.getenv('DETECTOR_MIN_W', '0'))
    min_h = int(os.getenv('DETECTOR_MIN_H', '0'))

    try:
        basic_check(image_path, min_w, min_h)
    except NudeCrawlerException as e:
        sys.exit(0)

    c = detect_nudity(image_path, address=detector_address, threshold=detector_threshold)
    sys.exit(c)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from flask import Flask, request
from nudenet import NudeClassifier
import os
import sys
import daemon
from daemon import pidfile
#import lockfile
import argparse
import signal
from PIL import UnidentifiedImageError


classifier = None

app = Flask(__name__)

pidfile_path = '/tmp/.nudenet-server.pid'

@app.get("/ping")
def ping():
    return "ну вот pong, и что?"


@app.route("/detect", methods=["POST"])
def detect():
    path = request.json['path']
    page = request.json['page']
    
    print("....", page)
    try:
        r = classifier.classify(path)
    except UnidentifiedImageError as e:
        print(f"Err: {page} {e}")
        result = {
            'status': 'ERROR',
            'error': str(e)
        }
        return result
    except Exception as e:
        print(f"Got uncaught exception {type(e)}: {e}")
    

    # sometimes no exception, but empty response, e.g. when mp4 instead of image
    if not r:
        print(f"Err: {page} empty reply")
        result = {
            'status': 'ERROR',
            'error': "empty reply from classifier"
        }
        return result
    
    
    if r[path]['unsafe'] > 0.5:
        verdict="UNSAFE"
    else:
        verdict="safe"

    r[path]['status'] = 'OK'
    print(f'{verdict} ({r[path]["unsafe"]:.2f}) {page}')
    return r[path]


def get_args():
    parser = argparse.ArgumentParser('Daemonical REST API for NudeNet')
    parser.add_argument("--port", type=int, default=5000)
    parser.add_argument("-d", '--daemon', action='store_true', default=False)
    parser.add_argument('--kill', action='store_true', default=False)
    parser.add_argument('--pidfile', default=pidfile_path)
    return parser.parse_args()

def sanity_check():
    def print_help(msg):
        print(f"{msg}Download original file from https://nudecrawler.netlify.app/classifier_model.onnx\n" \
              "Or from https://github.com/notAI-tech/NudeNet", file=sys.stderr)
        
    min_sane_size = 10*1024*1024
    path = os.path.expanduser('~/.NudeNet/classifier_model.onnx')

    if not os.path.exists(path):
        print_help(f"Missing {path}\n")
        return False

    sz = os.stat(path).st_size
    
    # check if size is OK. Normally its 80M, we check if it's at least 10M
    if sz <= min_sane_size:
        print_help(f"Too small file ({sz}) {path}\n")
        return False
    
    return True


def main():
    global classifier
    global pidfile
    args = get_args()
    if not sanity_check():
        sys.exit(1)
    
    print("load classified")
    classifier = NudeClassifier()

    if args.kill:
        try:
            with open(pidfile_path) as fh:
                pid = int(fh.read())
                print("Killing nudenet server with pid", pid)
                os.kill(pid, signal.SIGINT)
            os.unlink(pidfile_path)
        except FileNotFoundError:
            print("no pidfile", pidfile_path, "not doing anything")
        sys.exit(0)

    if args.daemon:
        print("work as daemon...")
        with daemon.DaemonContext(
            # pidfile=lockfile.FileLock(args.pidfile)
            pidfile=pidfile.TimeoutPIDLockFile(pidfile_path)
            ):
            # pid = os.getpid()
            # with open(pidfile, "w+") as fh:
            #    print(pid, file=fh)
            print("daemon app.run")
            app.run(port=args.port)
            print("after app.run")
    else:
        app.run(port=args.port) 
    
    print("done.")   

if __name__ == '__main__':
    main()
    


#!/usr/bin/env python

import argparse
import datetime
import json
import time
import sys
import os
import shlex
import shutil
import subprocess
import logging


from dotenv import load_dotenv

import nudecrawler 
from .. import Page, Unbuffered, load
from ..page import  get_processed_images, context_fields
from ..version import __version__
from ..cache import cache
# from ..verbose import printv
from ..config import get_args
import nudecrawler.tgru

import transliterate.discover 

transliterate.discover.autodiscover()

stats = {
    'cmd': None,
    'filter': {
        'expr': 'True',
        'min_image_size': None,
        'min_total_images': 0,
        'min_content_length': None,
        'max_pictures': None,
        'image_extensions': None,        
        'max_errors': None
    },
    'uptime': 0,
    'urls': 0,
    'words': 0,
    'word': None,
    'last': {
        'url': None,
        'status': None,
        'detailed': None,
    },
    'last_interesting': {
        'url': None,
        'status': None,
        'detailed': None,
    },
    

    'now': None,
    'processed_images': 0,
    'ignored_pages': 0,
    'found_interesting_pages': 0,
    'found_nude_images': 0,
    'found_new_nude_images': 0,
    'found_new_total_images': 0,
    'resume': dict(),
    'gap_max': 0,
    'gap_url': None,
    'cache_path': None,
    'cache_save': 1
}

previous_content_length = None

stats_file = None
stats_period = 60

stats_next_write = time.time() + stats_period

started = time.time()

logfile: str = None
stop_after = None
stop_each = None
refresh = None
detect_image = None
detect_url = None

#
# page_mintotal = 0

expr = 'True'

nude = 1
video = 1
verbose = False
all_found = True

filter_methods = {
    "true": ("builtin", ":true"),
    "false": ("builtin", ":false"),
    "mudepy": ("builtin", ":nude"),
    "nudenetb": ("builtin", ":nudenet"),
    "aid": ("image", "detect-image-aid"),
    "nsfwapi": ("image", "detect-image-nsfw-api"),
    "nudenet": ("image", "detect-image-nudenet")
}




def analyse(url):

    global stop_after, previous_content_length

    p = Page(url, all_found=all_found,
            detect_url=detect_url, detect_image=detect_image, ignore_content_length=previous_content_length,
            min_images_size=stats['filter']['min_image_size'], 
            image_extensions = stats['filter']['image_extensions'],
            min_total_images=stats['filter']['min_total_images'],
            max_errors=stats['filter']['max_errors'],
            max_pictures=stats['filter']['max_pictures'],
            expr = stats['filter']['expr'], min_content_length=stats['filter']['min_content_length'])

    stats['urls'] += 1
    
    p.check_all()
    
    stats['last']['url'] = url
    stats['last']['status'] = p._status
    stats['last']['detailed'] = p._status_detailed

    stats['found_new_total_images'] += p.new_total_images
    stats['found_new_nude_images'] += p.new_nude_images

    previous_content_length = p.content_length

    if p.status().startswith('INTERESTING'):        
        stats['found_interesting_pages'] += 1
        stats['found_nude_images'] += p.nude_images
        stats['last_interesting']['url'] = url
        stats['last_interesting']['status'] = p._status
        stats['last_interesting']['detailed'] = p._status_detailed

        if logfile:
            with open(logfile, "a") as fh:
                if logfile.endswith('.json') or logfile.endswith('.jsonl'):
                    print(p.as_json(), file=fh)
                else:
                    print(p, file=fh)
    
    if p.status().startswith("INTERESTING") or verbose:
        print(p)

    if p.status().startswith("IGNORED"):
        stats['ignored_pages'] += 1

    save_stats(force=False)

    if stats['cache_path']:
        cache.save_conditional(stats['cache_path'], stats['cache_save'])

    if stop_after is not None and get_processed_images() > stop_after:
        print("Stop/refresh after processed", get_processed_images(), "images...")
        if refresh:
            # print("Refresh:", refresh)
            subprocess.run(refresh)

            # schedule next stop
            stop_after = get_processed_images() + stop_each
        else:
            print("No --refresh, exiting with code 2")
            sys.exit(2)

    return p


def save_stats(force=False):
    global stats_next_write    

    if stats_file is None:
        return

    if time.time() > stats_next_write or force:
        stats['now'] = datetime.datetime.now().strftime("%m/%d/%Y %H:%M:%S")
        stats['uptime'] = int(time.time() - started)
        stats['processed_images'] = get_processed_images()
        
        stats['cache'] = cache.status()

        with open(stats_file, "w") as fh:
            json.dump(stats, fh, indent=4)
            stats_next_write = time.time() + stats_period

    


def check_word(word, day, fails, 
               resumecount=None):

    global previous_content_length

    word = word.replace(' ','-').translate({ord('ь'): '', ord('ъ'): ''})

    if word.startswith("https://"):
        baseurl = word
    else:
        trans_word = transliterate.translit(word, 'tgru', reversed=True)
        baseurl=f'https://telegra.ph/{trans_word}'


    stats['word'] = word
    stats['words'] += 1

    url=f'{baseurl}-{day.month:02}-{day.day:02}'

    stats['resume']['month'] = day.month
    stats['resume']['day'] = day.day    
    stats['resume']['count'] = resumecount

    previous_content_length = None


    # r = requests.get(url)  
    if not resumecount:
        p = analyse(url)
        #if p.ignore:
        #    return
        c=2
    else:
        c=resumecount
        print(f"Resume from word {word} count {c}")
    
    

    nfails=0    
    while nfails<fails:        
        url=f'{baseurl}-{day.month:02}-{day.day:02}-{c}'        
        p = analyse(url)

        if p.http_code == 404:
            nfails += 1
        else:
            # end of gap
            if nfails>stats['gap_max']:
                stats['gap_max'] = nfails
                stats['gap_url'] = url
            nfails=0

        c+=1
        stats['resume']['count'] = c


def sanity_check(args):
    pass

def load_stats(path):
    global stats
    with open(path) as fh:
        loaded_stats = json.load(fh)
    
    for k in stats:
        if k not in loaded_stats:
            loaded_stats[k] = stats[k]
    
    stats = loaded_stats

def abort(msg):
    print(msg, file=sys.stderr)
    sys.exit(1)

def main():
    global nude, video, verbose, all_found, stats_file, stats, logfile, \
        stop_after, stop_each, detect_image, detect_url,\
        refresh

    words = None
    args = get_args(argv=None, methods_list=', '.join(filter_methods.keys()), context_fields=context_fields)
    sanity_check(args)

    # when fastforward, we go to specific word/day/count quickly
    fastforward = False

    # Disabled logger for nudenet
    logging.getLogger().disabled = True

    if args.unbuffered:
        sys.stdout = Unbuffered(sys.stdout)

    if args.resume:
        if args.workdir:
            args.resume = os.path.join(args.workdir, args.resume)

        print("Resume from", args.resume)
        try:
            load_stats(args.resume)
        except FileNotFoundError as e: 
            abort(f"Missing status file {args.resume}")

        cmd = stats['cmd']
        args = get_args(argv = shlex.split(cmd)[1:])
        fastforward = True
    else:
        stats['cmd'] = shlex.join(sys.argv)

    if args.workdir:
        for attr in ['cache', 'wordlist', 'log', 'resume', 'stats']:
            old = getattr(args, attr)
            if old is not None:                
                new = os.path.join(args.workdir, old)
                setattr(args, attr, new)

    # nude = args.nude
    # video = args.video
    verbose = args.verbose
    all_found = args.all    
    matched_resume = False
    skipped_words = 0
    stop_after = args.stop
    stop_each = args.stop
    refresh = args.refresh
    detect_url = args.detect_url
    detect_image = args.detect_image
    stats['filter']['expr'] = args.expr
    stats['filter']['min_content_length'] = args.min_content_length
    stats['filter']['max_errors'] = args.max_errors
    stats['filter']['max_pictures'] = args.max_pictures
    stats['cache_path'] = args.cache
    stats['cache_save'] = args.cache_save    

    if args.detect:
        try:
            kind, basename = filter_methods[args.detect]
        except KeyError:
            print(f"Do not know detector {args.detect!r}, use one of known detectors: ({ ', '.join(filter_methods.keys()) }) or explicitly specify script with --detect-image or --detect-url")
            sys.exit(1)

        if kind in ['image', 'url']:
            if shutil.which(basename) is None:
                print(f"Cannot find {basename}, maybe not in $PATH?" ,file=sys.stderr)
                sys.exit(1)

        if kind == 'builtin':
            if basename in [':nude', ':nudenet']:
                detect_image = basename
            else:
                detect_url = basename
        elif kind == 'image':
            detect_image = basename
            print(f"# Will use script {shutil.which(basename)} for filtering images")
        elif kind == 'url':
            detect_url = basename
            print(f"# Will use script {shutil.which(basename)} for filtering images")            
    

    # fix arguments
    if not any([detect_image, detect_url, all_found]):        
        print("# No nudity detector (--detect, --detect-url, --detect-image) given, using built-in --detect-image :nude by default")
        detect_image=':nude'

    nudecrawler.verbose.verbose = verbose

    if args.extensions:
        stats['filter']['image_extensions'] = args.extensions
    
    if args.minsize:
        stats['filter']['min_image_size'] = args.minsize * 1024

    if args.total:
        stats['filter']['min_total_images'] = args.total


    if stats['cache_path']:
        if os.path.exists(stats['cache_path']):
            cache.load(stats['cache_path'])
        else:
            print(f"# No cache file {stats['cache_path']}, start with empty cache")

    # processing could start here

    nudecrawler.load()

    # --url1 
    if args.url1:
        p = analyse(args.url1)
        print(p.status())
        for msg in p._log:
            print(" ", msg)
        return

    ## wordlist
    if args.wordlist:
        stats_file = args.stats
        with open(args.wordlist) as fh:
            words = [line.rstrip() for line in fh]
    
    if args.words:
        words = args.words
    
    if not words:
        print("Need either --url1 URL or words like 'nude' or -w wordlist.txt")
        sys.exit(1)

    logfile = args.log

    for w in words:
        if fastforward and not matched_resume:
            if w == stats['resume']['word']:
                matched_resume = True
            else:
                skipped_words += 1
                continue

        stats['resume']['word'] = w


        if fastforward:
            day = datetime.datetime(2020, stats['resume']['month'], stats['resume']['day'])
        elif args.day is None:
            day = datetime.datetime.now()
        else:
            day = datetime.datetime(2020, args.day[0], args.day[1])

        days_tried = 0
        while days_tried < args.days:
            if fastforward:
                resumecount = stats['resume']['count']
            else:
                resumecount = None
            # stop fastforward
            fastforward=False
            check_word(w, day, args.fails, resumecount=resumecount)
            
            days_tried += 1
            day = day - datetime.timedelta(days=1)


    print(f"Finished {len(words)} (skipped {skipped_words}) words in {time.time() - started:.2f} seconds, found {stats['found_interesting_pages']} pages")
    if fastforward and not matched_resume:
        abort(f"Did not found word {stats['resume']['word']}")

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt as e:
        print("KEYBOARD INTERRUPT")
        print(e)
        save_stats(force=True)
        


import inspect
import os
import requests
import json
import base64

verbose = False
# send_bugreports = False

def get_verbose():
    return verbose

def printv(*args):
    if not verbose:
        return
    
    if False:
        frame = inspect.stack()[1]
        location = os.path.basename(frame.filename) + ':' + str(frame.lineno)
        print(f"...", f"({location})", *args)
    else:
        print(f"...", *args)





from bs4 import BeautifulSoup
import urllib.request
from urllib.parse import urljoin
from .remoteimage import RemoteImage
from . import verbose
from .verbose import printv
from .exceptions import *
from .cache import cache

from evalidate import Expr, EvalException

import requests
import hashlib
import http
import time
import json


from urllib.parse import urlparse
import subprocess
import sys
import os

#def trivial_iterator(xx):
#    for x in xx:
#        yield x

processed_images = 0
context_fields = ['total_images', 'nude_images', 'nonnude_images', 'new_nude_images', 'new_nonnude_images', 'new_total_images', 'total_video']

def get_processed_images():
    return processed_images

def sha1sum(path):
    sum = hashlib.sha1()
    with open(path, 'rb') as source:
        block = source.read(2**16)
        while len(block) != 0:
            sum.update(block)
            block = source.read(2**16)
    return sum.hexdigest()


class Page:

    def __init__(self, url: str, all_found=False, detect_image=None, min_total_images = 0, min_images_size = 10*1024,
                 image_extensions=None, max_errors = None, max_pictures=None,
                 detect_url=None, min_content_length=None, ignore_content_length=None, expr='True'):
        self.url = url
        self.nban_links = 0
        self.nban_images = 0
        self.nude_images = 0
        self.nonnude_images = 0        
        self.total_images = 0

        # images not found in cache
        self.new_nude_images = 0
        self.new_nonnude_images = 0        
        self.new_total_images = 0

        self.total_video = 0
        self.text_found = list()
        self.links = list()        
        self.images = list()
        self.videos = list()
        self.all_found = all_found
        self.ignore_content_length = ignore_content_length
        
        
        # minor errors
        self.error_counter = 0

        self.detect_image = detect_image
        self.detect_url = detect_url
        self.image_extensions = image_extensions or ['.jpg', '.jpeg']
        self.min_image_size = min_images_size
        self.min_total_images = min_total_images
        self.max_errors = max_errors
        self.max_pictures = max_pictures

        # expr to filter interesing
        self._code = None

        self.http_code = None

        self._ignore = False # Ignore this page, we think it's spam, duplicate        
        self._status = None
        self._status_detailed = None
        self._status_logged = False
        self._log = list()
        self.check_time = None
        self.content_length = None

        # can throw evalidate.EvalExpression here
        # node = Expr(expr).code
        # self._code = compile(node, '<user filter>', 'eval')
        self._code = Expr(expr).code
        
        printv("Processing:", self.url)

        try:
            page = urllib.request.urlopen(self.url)
            self.http_code = page.getcode()
            self.content_length = page.headers.get('content-length')
            if self.content_length is not None:
                self.content_length = int(self.content_length)

            if self.content_length and min_content_length and self.content_length < min_content_length:
                self.ignore(f"content-length {self.content_length} < minimal {min_content_length}")
                return
        except (urllib.error.URLError, ConnectionError, http.client.RemoteDisconnected) as e:
            if hasattr(e, 'status') and e.status == 404:
                # print(e, type(e))
                # silent ignore most usual error (unless verbose)
                printv(url, 404)
                self._status = "IGNORED"
                self._status_detailed = "404"
                self._ignore = True
                self.http_code = e.status
            else:
                if hasattr(e, 'status'):
                    self.http_code = e.status
                self.ignore(f'Exception {e} with {self.url}')            
            return
        self.content_length = page.headers.get('content-length')
        
        if self.ignore_content_length is not None and self.ignore_content_length == self.content_length:
            self.ignore(f"Ignore because matches prev page content-length = {self.content_length}")
            return
        
        self.soup = BeautifulSoup(page, "html.parser")


    def ignore(self, reason):
        self._ignore = True        
        self._status = f"IGNORED"
        self._status_detailed = reason
        printv(f"IGNORE {self.url}, {self._status_detailed}")
        self.log(self._status_detailed)


    def log(self, msg, really=True):
        if not really:
            return        
        self._log.append(msg)

    def check_all(self):
        started = time.time()

        if self._ignore:            
            return
        
        self.total_images = len(self.soup.findAll('img'))
        self.total_links = len(self.soup.findAll('a'))

        self.check_images()
        self.check_video()
    
        self.check_time = round(time.time() - started, 2)
        self.log(f"Check time: {self.check_time}")
        # set status
        self.status()
        

    def check_video(self):
        self.total_video = len(self.soup.findAll('video'))
        for img in self.soup.findAll('video'):
            src = img.get('src')
            self.videos.append(src)

    def error(self, msg):
        self.error_counter += 1
        self.log(f"minor error ({self.error_counter}): {msg}")
        if self.max_errors is not None and self.error_counter > self.max_errors:
            self.ignore(f'Too many minor errors: {self.max_errors}')            

    def prefilter_image(self, url):
        """ True is we should check, False if we can ignore this image"""
        
        if self._ignore:
            return False
        verdict = cache.url2v(url)
        
        if verdict is not None:
            # self.log(f'{url} passed prefilter because cached')
            return True 
        
        path = urlparse(url).path
        ext = os.path.splitext(path)[1]
        if ext not in self.image_extensions:            
            self.log(f"{url} bad extension, ignore")
            return False
        try:
            r = requests.head(url, timeout=1)
        except requests.exceptions.RequestException as e:            
            self.error(f'{url} request exception: {e}')
            return False

        if r.status_code != 200:
            self.error(f"Bad status code: {url} {r.status_code}")
            return False
        try:
            cl = int(r.headers['Content-Length'])
        except KeyError:
            cl = None

        # self.log(f"{url} status:{r.status_code} content-length: {cl}")
        if cl is not None and cl < self.min_image_size:
            self.log(f"Too small image ({int(r.headers['Content-Length'])})")
            return False

        # self.log(f"{url} image passed prefilter")
        return True


    def do_detect_url(self, url):

        if self.detect_url == ':true':
            self.nude_images += 1
            return True
        
        if self.detect_url == ':false':                
            return False            

        rc = subprocess.run([self.detect_url, url])
        if rc.returncode >= 100:
            print("FATAL ERROR")
            sys.exit(1)
        if rc.returncode:
            self.log(f"{url} is nude")
            self.nude_images += 1
        else:
            self.log(f"{url} is NOT nude")
            self.nonnude_images += 1
        
        return 

    def do_detect_image(self, url):
        try:
            ri = RemoteImage(url)
            sum = sha1sum(ri.path)
            verdict = cache.sum2v(sum, url=url)
            if verdict is not None:                    
                if verdict:
                    self.log(f"{url} is nude (cached sum)")
                    self.nude_images += 1
                else:
                    self.log(f"{url} is NOT nude (cached sum)")
                    self.nonnude_images += 1
                return verdict

            verdict = ri.detect_image(self.detect_image)                
            self.new_total_images += 1

            if verdict:
                self.new_nude_images += 1
            else:
                self.new_nonnude_images += 1
            
            cache.register(url, sum, verdict)
            if verdict:
                self.log(f"{url} is nude")
                self.nude_images += 1
            else:
                self.log(f"{url} is NOT nude")                    
                self.nonnude_images += 1
            return 
        except NudeCrawlerException as e:
            print(e)
            return

    def detect_cache_url(self, url):
        verdict = cache.url2v(url)
        if verdict is not None:
            if verdict:
                self.log(f"{url} is nude (cached url)")
                self.nude_images += 1
            else:
                self.log(f"{url} is NOT nude (cached url)")
                self.nonnude_images += 1
        return verdict
        

    def is_nude(self, url):
        os.environ["NUDECRAWLER_PAGE_URL"] = self.url
        os.environ["NUDECRAWLER_IMAGE_URL"] = url

        if self.detect_cache_url(url) is not None:
            return

        if self.detect_url:
            self.do_detect_url(url)
            return

        if self.detect_image:
            try:
                self.do_detect_image(url)
            except Exception as e:
                printv("Broken image:", url)
                print(e)
            return
                
        
        raise(Exception('Need either detect_image or detect_url or all'))

    def check_images(self):
        global processed_images

        image_list = [ urljoin(self.url, img.get('src')) for img in self.soup.findAll('img') ]
        self.log(f"total images on page: {len(image_list)}")
        
        if len(image_list) < self.min_total_images:
            self.ignore(f'Skip because total images {len(image_list)} < {self.min_total_images}')
            return
        
        image_list = list(filter(None, [ url if self.prefilter_image(url) else None for url in image_list ]))

        if self._ignore:            
            return

        self.log(f"total prefiltered images on page: {len(image_list)}")
        if len(image_list) < self.min_total_images:            
            self.ignore(f'Skip because total prefiltered images {len(image_list)} < {self.min_total_images}')
            return

        if not self.all_found:
            for url in image_list[:self.max_pictures]:
                self.is_nude(url)
                processed_images += 1


    def status(self):        

        if not self._status_logged:
            self._status_logged = True
            makelog = True
        else:
            makelog = False
        
        if self._ignore:
            return f"{self._status} ({self._status_detailed})"

        self._status_detailed = f"total: {self.total_images} (min: {self.min_total_images}) " \
                    f"nude: {self.nude_images}"

        self._status_detailed += f" Cache new/nude: {self.new_total_images}/{self.new_nude_images}"

        if self.total_video:
            self._status_detailed += f" video: {self.total_video}"
        self._status_detailed += f" ({self.check_time}s)"
        
        self.log(self._status_detailed, makelog)

        if self.all_found:
            self._status = "INTERESTING (ALL)"
            self.log(self._status, makelog)        
            return self._status

        # Interesting or not? use evalidate
        ctx = dict()
        for field in context_fields :
            ctx[field] = getattr(self, field)
        r = eval(self._code, ctx.copy())
        if r:
            self._status = f'INTERESTING'
            self.log(self._status, makelog)        
            return self._status            


        self._status = "???"
        self.log(self._status, makelog)        
        return self._status            

    def as_json(self):
        j = dict()
        j['status'] = self.status()

        for attr in ['url', 'total_images', 'nude_images', 'new_nude_images', 'nonnude_images', 'new_nonnude_images', 'total_video']:
            j[attr] = getattr(self, attr)

        return json.dumps(j)

    def __str__(self):
        
        text = ''
        
        text += f'{self.status()} {self.url} ({self.check_time}s)\n'

        if self.all_found:
            text += f"  Total images: {self.total_images}\n"
        else:
            text += f"  Nude: {self.nude_images} ({self.new_nude_images} new) non-nude: {self.nonnude_images} ({self.new_nonnude_images} new)\n"
        
        #if self.new_nude_images or self.new_nonnude_images:
        #    text += f"  New nude {self.new_nude_images} non-nude {self.new_nonnude_images}\n"
                
        if self.total_video:
            text += f"  Total video: {self.total_video}\n"
        
        return text




import json
import os
import sys
from .verbose import printv

class ImageCache(object):
    def __init__(self):
        self._sum2v = dict()
        self._url2sum = dict()
        
        self.miss_url = 0
        self.miss_sum = 0
        self.hit_url = 0
        self.hit_sum = 0
        self._new = 0
            
    def url2v(self, url):
        try:
            sum = self._url2sum[url]
            v = self._sum2v[sum]
            self.hit_url +=1
            return v
        except KeyError:
            self.miss_url += 1
            return None            

    def sum2v(self, sum, url=None):
        try:
            v = self._sum2v[sum]
            self.hit_sum += 1
            # if we are here, url (most likely) not in _url2v
            if url:
                self._url2sum[url] = sum
                self._new += 1
            return v
        except KeyError:
            self.miss_sum += 1
            return None
    
    def register(self, url, sum, verdict):
        self._url2sum[url] = sum
        self._sum2v[sum] = verdict
        self._new += 1


    def load(self, path):
        with open(path) as fh:
            try:
                cache = json.load(fh)
            except json.decoder.JSONDecodeError:
                print('Invalid JSON in cache file', path)
                print('Fix or delete file and restart')
                sys.exit(1)

        
        self._url2sum = cache['_url2sum']
        self._sum2v = cache['_sum2v']
        print(f"Loaded {len(self._url2sum)} urls and {len(self._sum2v)} sums cache")

    def save_conditional(self, path, new=1):
        tmppath = path + '.tmp'
        if self._new >= new:
            printv(f"Save cache with {self._new} updates")
            self.save(tmppath)
            os.rename(tmppath, path)

    def save(self, path):
        data = dict(_url2sum = self._url2sum, _sum2v = self._sum2v)
        with open(path, "w") as fh:
            json.dump(data, fh, indent=2)
        self._new = 0

    def status(self):
        return {
            'urls': len(self._url2sum),
            'sums': len(self._sum2v),
            'hit_url': self.hit_url,
            'hit_sum': self.hit_sum,
            'miss_url': self.miss_url,
            'miss_sum': self.miss_sum
        }
        

cache = ImageCache()



from PIL import Image, UnidentifiedImageError
from .exceptions import BoringImage, ProblemImage

def basic_check(path, min_w, min_h):
    """ check image and raise exception """
    
    if not (min_w or min_h):        
        return
    
    try:
        img = Image.open(path)
    except UnidentifiedImageError:
        raise ProblemImage('Incorrect image')
    w, h = img.size

    if min_w and w<min_w:
        raise BoringImage(f'Image {w}x{h} is too small (min_w: {min_w})')

    if min_h and w<min_h:
        raise BoringImage(f'Image {w}x{h} is too small (min_h: {min_h})')
    



import toml
import argparse
import os
from dotenv import load_dotenv
from .version import __version__
from rich.pretty import pprint
from collections import ChainMap
import collections.abc

def get_config_path():

    def first_file(file_paths):
        for file_path in file_paths:
            if not file_path:
                continue
            if os.path.isfile(file_path):
                return file_path
        return None

    config_path_list = [
        os.getenv('NUDECRAWLER_CONFIG'),
        'nudecrawler.toml',
        os.path.expanduser('~/nudecrawler.toml'),
        '/etc/nudecrawler.toml'
    ]    

    def_config_path = first_file(config_path_list)
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-c', '--config', default=def_config_path)
    args, _ = parser.parse_known_args()
    return args.config


def get_args(argv: str, methods_list: list[str], context_fields: list):

    load_dotenv()  

    config_path = get_config_path()

    try:
        assert config_path is not None
        with open(config_path, 'r') as f:
            config = toml.load(f)
    except AssertionError:
        pprint("No config files found")
        config = dict()

    config = hydrate_config(config)
    pprint(config)

    print(config["cache"]["cache"])

    parser = argparse.ArgumentParser(description=f'Nudecrawler: Telegra.ph Spider {__version__}\nhttps://github.com/yaroslaff/nudecrawler', formatter_class=argparse.RawTextHelpFormatter)

    # 

    
    parser.add_argument('words', nargs='*')
    parser.add_argument('--url1', metavar="URL", help='process only one url')
    parser.add_argument('--day', nargs=2, type=int, metavar=('MONTH', 'DAY'), help='Current date (default is today) example: --day 12 31')
    parser.add_argument('-c', '--config', default=config_path, help=f'Path to nudecrawler.toml, ({config_path})')
    parser.add_argument('-v', '--verbose', default=["verbose"], action='store_true', help='verbose')
    parser.add_argument('--unbuffered', '-b', default=config["unbuffered"], action='store_true', help='Use unbuffered stdout')


    g = parser.add_argument_group('Depth')
    g.add_argument('-d', '--days', type=int, default=config['depth']['days'], help=f'check N days back ({config["depth"]["days"]})')
    g.add_argument('-f', '--fails', type=int, default=config['depth']['fails'], help=f'stop searching next pages with same words after N failures ({config["depth"]["fails"]})')


    g = parser.add_argument_group('Page filtering options')
    g.add_argument('-a', '--all', default=False, action='store_true', help='do not detect, print all found pages')
    g.add_argument('--expr', '-e', metavar='EXPR', default=config['filter']['expr'], 
                        help=f'Interesting if EXPR is True. def: { config["filter"]["expr"] }\nFields: ' + ' '.join(context_fields) )
    g.add_argument('--total', metavar='N', type=int, default=config['filter']['total'], help=f'Skip detections if less then N total images ({config["filter"]["total"]})')
    g.add_argument('--max-errors', metavar='N', type=int, default=config['filter']['max_errors'], help=f'Max allowed errors on page ({config["filter"]["max_errors"]})')
    g.add_argument('--min-content-length', metavar='N', type=int, default=config['filter']['min-content-length'], help=f'Skip page if content-length less then N (try 5000 or higher)')

    g = parser.add_argument_group('Image detection options')
    g.add_argument('--detect', metavar='METHOD', default=config["detect"]["detect"], help=f'One of {methods_list} ({config["detect"]["detect"]})')
    g.add_argument('--extensions', nargs='*', default=config['detect']['extensions'],help='interesting extensions (with dot, like .jpg)')
    g.add_argument('--minsize', type=int, default=config['detect']['minsize'],help=f'min size of image in Kb ({config["detect"]["minsize"]})')
    g.add_argument('--max-pictures', type=int, metavar='N', default=config["detect"]["max-pictures"], help=f'Detect only among first prefiltered N pictures')


    g = parser.add_argument_group('Cache')
    g.add_argument('--cache', metavar='PATH', default=config["cache"]["cache"], help=f'path to cache file (will create if missing)')
    g.add_argument('--cache-save', type=int, metavar='N', default=config["cache"]["cache-save"], help=f'Save cache after N new images ({config["cache"]["cache-save"]})')

    g = parser.add_argument_group('Externat detector')
    g.add_argument('--detect-image', metavar='SCRIPT', default=config["detector"]["detect-image"], help='explicitly use this script to detect nudity on image file')
    g.add_argument('--detect-url', metavar='SCRIPT', default=config["detector"]["detect-url"], help='explicitly use this script to detect nudity on image URL')


    g = parser.add_argument_group('Long-run options')
    g.add_argument('--log', default=config["longrun"]["log"], help='print all precious treasures to this logfile')
    # g.add_argument('--bugreport', default=False, action='store_true', help='send bugreport in case of problem (works only after agreed in github ticket)')
    g.add_argument('--workdir', default=config["longrun"]["workdir"], help=f'Use all files (log, wordlist, cache) in this dir. def: { config["longrun"]["workdir"] }')

    g.add_argument('-w', '--wordlist', help='wordlist (urllist) file')
    g.add_argument('--stats', metavar='STATS_FILE', default=config["longrun"]["stats"], help='periodical statistics file')
    g.add_argument('--resume', metavar='STATS_FILE', help='resume from STATS_FILE (other args are not needed)')
    g.add_argument('--stop', type=int, default=config["longrun"]["stop"], metavar='NUM_IMAGES', help='stop (or --refresh) after N images processed (or little after)')
    g.add_argument('--refresh', metavar=('SCRIPT', 'ARG'), nargs='+', help='run this refresh script every --stop NUM_IMAGES images')

    return parser.parse_args(argv)

def hydrate_config(conf: dict):
    
    def update(d, u):
        """ recursive dict update """
        for k, v in u.items():
            if isinstance(v, collections.abc.Mapping):
                d[k] = update(d.get(k, {}), v)
            else:
                d[k] = v
        return d

    
    def_config = {
        'verbose': False,
        'unbuffered': False,
        'depth': {
            'days': 10,
            'fails': 5,
        },
        'filter': {
            'all': False,
            'expr': 'nude_images > 0',
            'total': int(os.getenv('NUDE_TOTAL', '1')),
            'max_errors': 5,
            'min-content-length': None
        },
        'detect': {
            'detect': os.getenv('NUDE_DETECT','mudepy'),
            'extensions': ['.jpeg','.jpg', '.png'],
            'minsize': 10,
            'max-pictures': None
        },
        'detector': {
            'detect-image': None,
            'detect-url': None,
        },
        'cache': {
            'cache': os.getenv('NUDE_CACHE'),
            'cache-save': 1000
        },
        'longrun': {
            'log': os.getenv('NUDE_LOG'),
            'workdir': os.getenv('NUDE_DIR', '.'),
            'wordlist': None,
            'stats': os.getenv('NUDE_STATS', '/tmp/nudecrawler-stats.txt'),
            'stop': None,
            'refresh': None
        }
    }

    return update(def_config, conf)




class Unbuffered(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, datas):
        self.stream.writelines(datas)
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)



from nudecrawler.page import Page
import nudecrawler.verbose 

empty = 'https://telegra.ph/empty-04-03'
belle_delphine = 'https://telegra.ph/belle-delphine-01-16'
sasha_grey = 'https://telegra.ph/sasha-grey-04-18'

nudecrawler.verbose.verbose = True

class TestBasic():

    def test_empty(self):
        p = Page(empty, detect_image=':nude')
        p.check_all()
        assert p.status().startswith('INTERESTING'), "Bad status!"
        print(p)

    def test_belle(self):
        p = Page(belle_delphine, detect_image=':true')
        p.check_all()
        assert p.status().startswith('INTERESTING'), "Bad status!"
        print(p)



class NudeCrawlerException(Exception):
    pass

class BoringImage(NudeCrawlerException):
    """ Image is boring, no need to analyse, e.g. small icon """
    pass

class ProblemImage(NudeCrawlerException):
    """ Technical problem with image, e.g. 404 NotFound or damaged file """
    pass




import PIL
from PIL import Image, UnidentifiedImageError
from pprint import pprint

import os
from urllib.parse import urlparse
import tempfile
import sys
import onnxruntime
import subprocess
# import nude
from .exceptions import ProblemImage
from .verbose import printv

# detector_address = 'http://localhost:9191/api/v1/detect'

nudenet_detector = None

def _load():
    global nudenet_classifier
    try:
        from nudenet import NudeDetector
        print("Loading nudenet classifier....")
        nudenet_detector = NudeDetector()
    except (ModuleNotFoundError, onnxruntime.capi.onnxruntime_pybind11_state.InvalidProtobuf):
        nudenet_detector = None



class RemoteImage:
    def __init__(self, url):        
        self.url = url
        self.path = None
        pr = urlparse(self.url)
        suffix = os.path.splitext(pr.path)[1]
        try:            
            r = requests.get(url, timeout=10)
        except requests.exceptions.RequestException as e:
            raise ProblemImage(f'Requests.exception for {self.url}: {e}')

        # r.raise_for_status()
        if r.status_code != 200:
            raise ProblemImage(f'Bad status {r.status_code} for {self.url}')

        self.threshold = float(os.getenv('NUDE_DETECT_THRESHOLD', '0.5'))

        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as tmp:
            self.path = tmp.name
            tmp.file.write(r.content)

    def set_threshold(self, thr):
        self.threshold = thr

    def __del__(self):
        if self.path:
            os.unlink(self.path)

    def download(self):
        pass

    def detect_image(self, script):
        """
        new method, using external script
        """

        if script == ':true':
            return True

        if script == ':false':
            return False

        if script==':nude':
            n = nude.Nude(self.path)            
            n.resize(maxheight=800, maxwidth=600)
            return n.parse().result
        
        if script==':nudenet':
            if nudenet_detector is None:
                print("built-in nudenet detector selected, but nudenet is not installed or model not loaded")
                sys.exit(1)
            return self.nudenet_detect()

        rc = subprocess.run([script, self.path], env=os.environ.copy())
        if rc.returncode >= 100:
            print("FATAL ERROR")
            sys.exit(1)        
        return bool(rc.returncode)

    def nudenet_detect(self):

        print("NUDENET DETECT")

        try:
            r = nudenet_detector.detect(self.path)
        except UnidentifiedImageError as e:
            print(f"Err: {self.url} {e}")
            result = {
                'status': 'ERROR',
                'error': str(e)
            }
            return False
        except Exception as e:
            print(f"Got uncaught exception {type(e)}: {e}")
        
        # sometimes no exception, but empty response, e.g. when mp4 instead of image
        if not r:
            printv(f"Err: {self.url} empty reply")
            return False

        pprint(r)

        if r[self.path]['unsafe'] > self.threshold:
            return True
        
        return False

